import pathlib
import re
import uuid
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import TypedDict

import phonenumbers
import tqdm
from phonenumbers import NumberParseException
from typer import Context as TyperContext
from typer import Argument
from typer import Option
from typer import Typer
from typer import echo

from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.data import ModelRow
from spinta.cli.helpers.data import count_rows
from spinta.cli.helpers.data import iter_model_rows
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import Node
from spinta.components import Property
from spinta.core.config import RawConfig
from spinta.dimensions.prefix.components import UriPrefix
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.manifests.tabular.helpers import write_tabular_manifest
from spinta.types.namespace import sort_models_by_refs
from spinta.utils.data import take
from spinta.utils.nin import is_nin_lt


app = Typer()


def _get_new_metadata_row_id():
    return str(uuid.uuid4())


PREFIXES = {
    'person': 'https://www.w3.org/ns/person#',
    'pii': 'https://data.gov.lt/pii/',
}


def _ensure_prefix(node: Node, uri: str) -> None:
    if isinstance(node, Property):
        manifest = node.model.manifest
    else:
        raise RuntimeError("Don't know how to get manifest from {node!r}.")

    prefix, _ = uri.split(':', 1)
    if prefix not in manifest.prefixes:
        prefix_ = UriPrefix()
        prefix_.id = _get_new_metadata_row_id()
        prefix_.name = prefix
        prefix_.uri = PREFIXES[prefix]
        manifest.prefixes[prefix] = prefix_


class PiiMatch(TypedDict):
    num: int    # Number of rows matches
    total: int  # Total number of rows checked


PiiMatches = Dict[
    str,            # model.name
    Dict[
        str,        # property.place
        Dict[
            str,    # PII URI in `prefix:suffix` form.
            PiiMatch,
        ],
    ]
]


def _add_pii_match(
    matches: PiiMatches,
    model: Model,
    prop: Property,
    pii_: str,
    matched: bool,
) -> None:
    if model.name not in matches:
        matches[model.name] = {}

    if prop.place not in matches[model.name]:
        matches[model.name][prop.place] = {}

    if pii_ not in matches[model.name][prop.place]:
        matches[model.name][prop.place][pii_] = {
            'num': 0,
            'total': 0,
        }

    matches[model.name][prop.place][pii_]['total'] += 1

    if matched:
        matches[model.name][prop.place][pii_]['num'] += 1


email_re = re.compile(r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)')


def _detect_email(value: Any) -> bool:
    if not isinstance(value, str):
        return False
    if '@' not in value:
        return False
    return email_re.match(value) is not None


def _detect_phone(value: Any) -> bool:
    if isinstance(value, str):
        try:
            phone = phonenumbers.parse(value, 'LT')
        except NumberParseException:
            return False
        else:
            return phonenumbers.is_valid_number(phone)
    return False


def _detect_nin_lt(value: Any):
    return is_nin_lt(str(value))


def _detect_pii(manifest: Manifest, rows: Iterable[ModelRow]) -> None:
    """Detects PII and modifies given manifest in place"""

    detectors = [
        (_detect_nin_lt, 'pii:id'),
        (_detect_email, 'pii:email'),
        (_detect_phone, 'pii:phone'),
    ]

    # Detect PII properties.
    result: PiiMatches = {}
    for model, row in rows:
        for prop in take(model.properties).values():
            if prop.name in row:
                value = row[prop.name]
                for detector, uri in detectors:
                    if _add_pii_match(
                        result,
                        model,
                        prop,
                        uri,
                        detector(value),
                    ):
                        break

    # Update manifest.
    for model_name, props in result.items():
        model = manifest.models[model_name]
        for prop_place, matches in props.items():
            prop = model.flatprops[prop_place]
            for uri, match in matches.items():
                percent = match['num'] / match['total'] * 100
                if percent > 50:
                    _ensure_prefix(prop, uri)
                    prop.uri = uri


@app.command()
def detect(
    ctx: TyperContext,
    manifest: Optional[pathlib.Path] = Argument(None, help="Path to manifest."),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Path to manifest with detected PII."
    )),
    auth: Optional[str] = Option(None, '-a', '--auth', help=(
        "Authorize as a client"
    )),
    stop_on_error: bool = Option(False, help="Stop on first error"),
    limit: Optional[int] = Option(1000, help="Limit number of rows to check"),
):
    """Detect Person Identifying Information

    This will inspect row values and checks if any of value detection functions
    returns true. For those columns having more than 50% of values matching the
    checks a uri value is set.
    """
    context: Context = ctx.obj

    if manifest:
        config = {
            'backends.cli': {
                'type': 'memory',
            },
            'keymaps.default': {
                'type': 'sqlalchemy',
                'dsn': 'sqlite:///{data_dir}/keymap.db',
            },
            'manifests.cli': {
                'type': 'tabular',
                'path': str(manifest),
                'backend': 'cli',
                'keymap': 'default',
                'mode': 'external',
            },
            'manifest': 'cli',
        }

        # Add given manifest file to configuration
        rc: RawConfig = context.get('rc')
        context: Context = context.fork('detect')
        context.set('rc', rc.fork(config))

    # Load manifest
    store = prepare_manifest(context, verbose=False)
    manifest = store.manifest
    with context:
        require_auth(context, auth)
        context.attach('transaction', manifest.backend.transaction)
        backends = set()
        for backend in store.backends.values():
            backends.add(backend.name)
            context.attach(f'transaction.{backend.name}', backend.begin)
        for backend in manifest.backends.values():
            backends.add(backend.name)
            context.attach(f'transaction.{backend.name}', backend.begin)
        for dataset in manifest.datasets.values():
            for resource in dataset.resources.values():
                if resource.backend and resource.backend.name not in backends:
                    backends.add(resource.backend.name)
                    context.attach(f'transaction.{resource.backend.name}', resource.backend.begin)
        for keymap in store.keymaps.values():
            context.attach(f'keymap.{keymap.name}', lambda: keymap)

        from spinta.types.namespace import traverse_ns_models

        ns = manifest.objects['ns']['']
        models = traverse_ns_models(context, ns, Action.SEARCH)
        models = sort_models_by_refs(models)
        models = list(reversed(list(models)))
        counts = count_rows(context, models, limit=limit)

        rows = iter_model_rows(
            context, models, counts,
            stop_on_error=stop_on_error,
            limit=limit,
        )
        total = sum(counts.values())
        rows = tqdm.tqdm(rows, 'PII DETECT', ascii=True, total=total)
        _detect_pii(manifest, rows)
        if output:
            write_tabular_manifest(output, manifest)
        else:
            echo(render_tabular_manifest(manifest))



from typing import Any
from typing import Dict

from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.frictionless.components import FrictionlessBackend


@commands.load.register(Context, FrictionlessBackend, dict)
def load(context: Context, backend: FrictionlessBackend, config: Dict[str, Any]):
    dsn = config['dsn']



import hashlib
import os


def get_new_id(resource_type, random_field=None):
    type_field = hashlib.sha512(resource_type.encode('UTF-8')).hexdigest()[:4]
    random_field = random_field or os.urandom(16).hex()
    checksum_field = hashlib.sha512((type_field + random_field).encode()).hexdigest()[:8]
    return '{}-{}-{}'.format(type_field, random_field, checksum_field)



from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Array
from spinta.migrations import SchemaVersion
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, type(None), Array)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: type(None),
    current: Array,
):
    list_table_name = get_pg_name(get_table_name(current.prop, TableType.LIST))
    main_table_name = get_pg_name(get_table_name(current.prop.model))

    props = [
        {
            'name': 'column',
            'args': ['_txn', {'name': 'uuid', 'args': []}],
        },
        {
            'name': 'column',
            'args': ['_rid', {'name': 'ref', 'args': [
                f'{main_table_name}._id', {
                    'name': 'bind',
                    'args': ['ondelete', 'CASCADE'],
                }
            ]}],
        },
    ]
    props += commands.freeze(context, version, backend, None, current.items.dtype)
    version.actions.append({
        'type': 'schema',
        'upgrade': {
            'name': 'create_table',
            'args': [list_table_name] + props,
        },
        'downgrade': {
            'name': 'drop_table',
            'args': [list_table_name],
        },
    })

    if current.prop.list is None:
        # For fast whole resource access we also store whole list in a JSONB.
        return [
            {
                'name': 'column',
                'args': [
                    current.prop.place,
                    {
                        'name': 'json',
                        'args': [],
                    },
                ]
            },
        ]

    else:
        return []


@commands.freeze.register(Context, SchemaVersion, PostgreSQL, Array, Array)
def freeze(
    context: Context,
    version: SchemaVersion,
    backend: PostgreSQL,
    freezed: Array,
    current: Array,
):
    commands.freeze(context, version, backend, freezed.items.dtype, current.items.dtype)



from __future__ import annotations

import collections
import inspect
import pathlib
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import cast

from multipledispatch.dispatcher import Dispatcher
from multipledispatch.dispatcher import str_signature

F = TypeVar('F', bound=Callable[[], Any])

_commands: Dict[str, Command] = {}


def command(
    name: str = None,
    schema: dict = None,
) -> Callable[[F], Command]:
    """Define a new command."""

    def _(func: F) -> Command:
        _name = func.__name__ if name is None else name
        if _name in _commands:
            raise Exception(f"{_name!r} is already defined.")
        _commands[_name] = Command(_name)
        _commands[_name].schema = schema or {}
        return _commands[_name]

    return cast(Command, _)


class Command(Dispatcher):
    schema: dict

    def register(
        self,
        *signature_types,
        schema=None,
    ) -> Callable[[F], Command]:
        def _(func: F) -> Command:
            types = signature_types or tuple(_find_func_types(func))
            self.add(types, func)
            return self
        return cast(Command, _)

    def __getitem__(self, types):
        types = types if isinstance(types, tuple) else (types,)
        func = self.dispatch(*types)
        if not func:
            raise NotImplementedError(
                f"Could not find signature for {self.name}: "
                f"<{str_signature(types)}>"
            )
        return func

    def print_methods(self, *args, **kwargs) -> None:
        """Print all commands method in resolution order."""
        if args:
            # Find method by given args.
            args = tuple([type(arg) for arg in args])
            func = self.dispatch(*args)
        else:
            func = None

        base = pathlib.Path().resolve()
        arg_names = _extend_duplicate_names(self.ordering)
        print('---')
        for args in self.ordering:
            func_ = self.funcs[args]
            mark = func_ is func
            _print_method(base, func_, self.name, arg_names, args, mark)


def _extend_duplicate_names(
    argslist: List[Tuple[type]]
) -> Dict[type, str]:

    argnames = collections.defaultdict(set)
    for args in argslist:
        for arg in args:
            name = arg.__name__
            argnames[name].add(arg)

    for i in range(10):
        found = False
        for name in list(argnames):
            if len(argnames[name]) == 1:
                continue
            found = True
            n = name.count('.') + 1
            args = argnames.pop(name)
            for arg in args:
                name = arg.__module__.split('.')[-n:] + [arg.__name__]
                name = '.'.join(name)
                argnames[name].add(arg)
        if not found:
            break

    return {next(iter(v)): k for k, v in argnames.items()}


def _print_method(base, func, name, argnames, args, mark=False):
    file = inspect.getsourcefile(func)
    line = inspect.getsourcelines(func)[1]
    try:
        file = pathlib.Path(file).relative_to(base)
    except ValueError:
        # If two paths do not have common base, then fallback to full
        # file path.
        pass
    argsn = ', '.join([argnames[arg] for arg in args])
    signature = f'{name}({argsn}):'
    marker = ' * ' if mark else '   '
    print(f'{marker}{signature:<60} {file}:{line}')


def _find_func_types(func):
    sig = inspect.signature(func)

    kinds = {
        inspect.Parameter.POSITIONAL_ONLY,
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
    }
    for param in sig.parameters.values():
        if param.kind not in kinds or param.annotation is param.empty:
            break
        yield param.annotation



from spinta import commands
from spinta.backends.fs.components import FileSystem
from spinta.types.file.helpers import prepare_patch_data
from spinta.utils.data import take
from spinta.types.datatype import File
from spinta.components import Context, DataSubItem, Action
from spinta.backends.mongo.components import Mongo


@commands.before_write.register(Context, File, Mongo)
def before_write(
    context: Context,
    dtype: File,
    backend: Mongo,
    *,
    data: DataSubItem,
):
    content = take('_content', data.patch)
    if isinstance(content, bytes) and isinstance(dtype.backend, FileSystem):
        filepath = dtype.backend.path / take('_id', data.given, data.saved)
        with open(filepath, 'wb') as f:
            f.write(content)

    patch = prepare_patch_data(dtype, data)

    if data.root.action == Action.INSERT:
        return {dtype.prop.place: patch}

    return {
        f'{dtype.prop.place}.{k}': v for k, v in patch.items()
    }



import sqlalchemy as sa

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import ExternalRef
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name


@commands.prepare.register(Context, PostgreSQL, ExternalRef)
def prepare(context: Context, backend: PostgreSQL, dtype: ExternalRef):
    if dtype.model.given.pkeys or dtype.explicit:
        columns = []
        for prop in dtype.refprops:
            original_place = prop.place
            prop.place = f"{dtype.prop.place}.{prop.place}"
            column = commands.prepare(
                context,
                backend,
                prop.dtype,
            )
            if isinstance(column, list):
                columns.extend(column)
            elif column is not None:
                columns.append(column)
            prop.place = original_place
        return columns

    else:
        column_name = get_column_name(dtype.prop) + '._id'
        column_type = commands.get_primary_key_type(context, backend)
        return sa.Column(
            column_name,
            column_type,
            nullable=not dtype.required,
            unique=dtype.unique,
        )



from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Env
from spinta.core.ufuncs import Negative
from spinta.core.ufuncs import Pair
from spinta.core.ufuncs import Positive
from spinta.core.ufuncs import ufunc


@ufunc.resolver(Env, str)
def bind(env: Env, name: str) -> Bind:
    return Bind(name)


@ufunc.resolver(Env, str, object)
def bind(env: Env, name: str, value: object) -> Pair:
    return Pair(name, value)


@ufunc.resolver(Env, str)
def negative(env: Env, name: str) -> Negative:
    return Negative(name)


@ufunc.resolver(Env, str)
def positive(env: Env, name: str) -> Positive:
    return Positive(name)


@ufunc.resolver(Env, Bind)
def negative(env: Env, bind_: Bind) -> Negative:
    return Negative(bind_.name)


@ufunc.resolver(Env, Bind)
def positive(env: Env, bind_: Bind) -> Positive:
    return Positive(bind_.name)



from copy import copy
from typing import Any, List
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Iterator
from typing import Optional
from typing import Tuple
from typing import TypeVar

from typer import Argument
from typer import Context as TyperContext
from typer import Option
from typer import echo

from spinta import commands
from spinta.cli.helpers.store import load_manifest
from spinta.components import Context, Property, Node, Namespace
from spinta.components import Mode
from spinta.core.config import RawConfig
from spinta.core.config import ResourceTuple
from spinta.core.config import parse_resource_args
from spinta.core.context import configure_context
from spinta.datasets.components import Dataset, Resource, Attribute, Entity, ExternalBackend
from spinta.dimensions.prefix.components import UriPrefix
from spinta.manifests.components import Manifest
from spinta.manifests.components import ManifestPath
from spinta.manifests.helpers import get_manifest_from_type, init_manifest
from spinta.manifests.tabular.helpers import render_tabular_manifest
from spinta.manifests.tabular.helpers import write_tabular_manifest
from spinta.types.datatype import Ref, DataType, Array, Object, Denorm
from spinta.utils.naming import Deduplicator
from spinta.utils.schema import NA
from spinta.utils.schema import NotAvailable
from spinta.components import Model


def inspect(
    ctx: TyperContext,
    manifest: Optional[str] = Argument(None, help="Path to manifest."),
    resource: Optional[Tuple[str, str]] = Option(
        (None, None), '-r', '--resource',
        help=(
            "Resource type and source URI "
            "(-r sql sqlite:////tmp/db.sqlite)"
        ),
    ),
    formula: str = Option('', '-f', '--formula', help=(
        "Formula if needed, to prepare resource for reading"
    )),
    backend: Optional[str] = Option(None, '-b', '--backend', help=(
        "Backend connection string"
    )),
    output: Optional[str] = Option(None, '-o', '--output', help=(
        "Output tabular manifest in a specified file"
    )),
    auth: Optional[str] = Option(None, '-a', '--auth', help=(
        "Authorize as a client"
    )),
    priority: str = Option('manifest', '-p', '--priority', help=(
        "Merge property priority ('manifest' or 'external')"
    ))
):
    """Update manifest schema from an external data source"""
    if priority not in ['manifest', 'external']:
        echo(f"Priority \'{priority}\' does not exist, there can only be \'manifest\' or \'external\', it will be set to default 'manifest'.")
        priority = 'manifest'
    has_manifest_priority = priority == 'manifest'
    resources = parse_resource_args(*resource, formula)
    context = configure_context(
        ctx.obj,
        [manifest] if manifest else None,
        mode=Mode.external,
        backend=backend,
    )
    store = load_manifest(context, ensure_config_dir=True)
    old = store.manifest
    manifest = Manifest()
    init_manifest(context, manifest, 'inspect')
    commands.merge(context, manifest, manifest, old, has_manifest_priority)

    if not resources:
        resources = []
        for ds in old.datasets.values():
            for resource in ds.resources.values():
                external = resource.external
                if external == '' and resource.backend:
                    external = resource.backend.config['dsn']
                if not any(res.external == external for res in resources):
                    resources.append(ResourceTuple(type=resource.type, external=external, prepare=resource.prepare))

    if resources:
        for resource in resources:
            _merge(context, manifest, manifest, resource, has_manifest_priority)

    # Sort models for render
    sorted_models = {}
    for key, model in manifest.models.items():
        if key not in sorted_models.keys():
            if model.external and model.external.resource:
                resource = model.external.resource
                for resource_key, resource_model in resource.models.items():
                    if resource_key not in sorted_models.keys():
                        sorted_models[resource_key] = resource_model
            else:
                sorted_models[key] = model
    manifest.objects['model'] = sorted_models

    if output:
        write_tabular_manifest(output, manifest)
    else:
        echo(render_tabular_manifest(manifest))


def _merge(context: Context, manifest: Manifest, old: Manifest, resource: ResourceTuple, has_manifest_priority: bool):
    rc: RawConfig = context.get('rc')
    Manifest_ = get_manifest_from_type(rc, resource.type)
    path = ManifestPath(type=Manifest_.type, path=resource.external)
    context = configure_context(context, [path], mode=Mode.external)
    store = load_manifest(context)
    new = store.manifest
    commands.merge(context, manifest, old, new, has_manifest_priority)


@commands.merge.register(Context, Manifest, Manifest, Manifest, bool)
def merge(context: Context, manifest: Manifest, old: Manifest, new: Manifest, has_manifest_priority: bool) -> None:
    resource_list = []
    for ds in new.datasets.values():
        for res in ds.resources.values():
            resource_list.append(_resource_source_key(res))

    backends = zipitems(
        old.backends.values(),
        new.backends.values(),
        _backend_dsn
    )

    backend_deduplicator = Deduplicator("_{}")
    for backend in backends:
        for o, n in backend:
            if not (o and n):
                if o:
                    backend_deduplicator(o.name)
                if n:
                    name = backend_deduplicator(n.name)
                    n.name = name
            merge(context, manifest, o, n)
    datasets = zipitems(
        old.datasets.values(),
        new.datasets.values(),
        _dataset_resource_source_key,
    )

    for ds in datasets:
        for o, n in ds:
            commands.merge(context, manifest, o, n, has_manifest_priority)


@commands.merge.register(Context, Manifest, NotAvailable, ExternalBackend)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: ExternalBackend) -> None:
    manifest.backends[new.name] = new


@commands.merge.register(Context, Manifest, ExternalBackend, ExternalBackend)
def merge(context: Context, manifest: Manifest, old: ExternalBackend, new: ExternalBackend) -> None:
    manifest.backends[old.name] = old


@commands.merge.register(Context, Manifest, ExternalBackend, NotAvailable)
def merge(context: Context, manifest: Manifest, old: ExternalBackend, new: NotAvailable) -> None:
    manifest.backends[old.name] = old


@commands.merge.register(Context, Manifest, NotAvailable, Dataset, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Dataset, has_manifest_priority: bool) -> None:
    manifest.datasets[new.name] = new
    _merge_resources(context, manifest, old, new)

    dataset_models = _filter_models_for_dataset(new.manifest, new)
    deduplicator = Deduplicator()
    for model in dataset_models:
        model.name = deduplicator(model.name)
        merge(context, manifest, NA, model, has_manifest_priority)

    new.manifest = manifest


@commands.merge.register(Context, Manifest, Dataset, Dataset, bool)
def merge(context: Context, manifest: Manifest, old: Dataset, new: Dataset, has_manifest_priority: bool) -> None:
    old.id = coalesce(old.id, new.id)
    old.description = coalesce(old.description, new.description)
    old.lang = coalesce(old.lang, new.lang)
    old.name = coalesce(old.name, new.name)
    old.manifest = coalesce(old.manifest, new.manifest)
    old.website = coalesce(old.website, new.website)
    old.projects = coalesce(old.projects, new.projects)
    old.source = coalesce(old.source, new.source)
    old.owner = coalesce(old.owner, new.owner)
    old.given = coalesce(old.given, new.given)
    old.level = coalesce(old.level, new.level)
    old.access = coalesce(old.access, new.access)
    old.title = coalesce(old.title, new.title)
    old.resources = coalesce(old.resources, new.resources)
    old.prefixes = coalesce(old.prefixes, new.prefixes)
    if old.ns and new.ns and old.ns.name == new.ns.name:
        commands.merge(context, manifest, old.ns, new.ns)
    else:
        old.ns = coalesce(old.ns, new.ns)
    manifest.datasets[old.name] = old
    _merge_prefixes(context, manifest, old, new)
    _merge_resources(context, manifest, old, new)

    dataset_models = _filter_models_for_dataset(manifest, old)
    models = zipitems(
        dataset_models,
        new.manifest.models.values(),
        _model_source_key
    )
    resource_list = []
    for res in new.resources.values():
        resource_list.append(_resource_source_key(res))
    deduplicator = Deduplicator()
    for model in models:
        for om, nm in model:
            if om:
                deduplicator(om.name)
            if om and not nm:
                if om.external and om.external.resource:
                    if _resource_source_key(om.external.resource) in resource_list:
                        om.external.name = None
            if not om and nm:
                if nm.external:
                    nm.external.dataset = old
                    name = deduplicator(f"{nm.external.dataset.name}/{nm.basename}")
                    nm.name = name
            merge(context, manifest, om, nm, has_manifest_priority)


@commands.merge.register(Context, Manifest, Dataset, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Dataset, new: NotAvailable, has_manifest_priority: bool) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, UriPrefix)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: UriPrefix) -> None:
    dataset = new.parent
    manifest.datasets[dataset.name].prefixes[new.name] = new


@commands.merge.register(Context, Manifest, UriPrefix, UriPrefix)
def merge(context: Context, manifest: Manifest, old: UriPrefix, new: UriPrefix) -> None:
    old.type = coalesce(new.type, old.type)
    old.id = coalesce(new.id, old.id)
    old.uri = coalesce(new.uri, old.uri)
    old.title = coalesce(new.title, old.title)
    old.description = coalesce(new.description, old.description)

    old.name = coalesce(old.name, new.name)


@commands.merge.register(Context, Manifest, UriPrefix, NotAvailable)
def merge(context: Context, manifest: Manifest, old: UriPrefix, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Namespace)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Namespace) -> None:
    manifest.namespaces[new.name] = new


@commands.merge.register(Context, Manifest, Namespace, Namespace)
def merge(context: Context, manifest: Manifest, old: Namespace, new: Namespace) -> None:
    old.keymap = coalesce(new.keymap, old.keymap)
    old.backend = coalesce(new.backend, old.backend)
    old.parent = coalesce(new.parent, old.parent)
    old.description = coalesce(new.description, old.description)
    old.generated = coalesce(new.generated, old.generated)
    old.lang = coalesce(new.lang, old.lang)
    old.enums = coalesce(new.enums, old.enums)

    old.name = coalesce(old.name, new.name)
    old.given = coalesce(old.given, new.given)
    old.title = coalesce(old.title, new.title)
    old.models = coalesce(old.models, new.models)
    old.access = coalesce(old.access, new.access)


@commands.merge.register(Context, Manifest, Namespace, NotAvailable)
def merge(context: Context, manifest: Manifest, old: Namespace, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Resource)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Resource) -> None:
    manifest.datasets[new.dataset.name].resources[new.name] = new


@commands.merge.register(Context, Manifest, Resource, Resource)
def merge(context: Context, manifest: Manifest, old: Resource, new: Resource) -> None:
    old.type = coalesce(new.type, old.type)
    old.description = coalesce(new.description, old.description)
    old.comments = coalesce(new.comments, old.comments)
    old.lang = coalesce(new.lang, old.lang)

    old.backend = coalesce(old.backend, new.backend)
    old.dataset = coalesce(old.dataset, new.dataset)
    old.models = old.models
    old.eid = coalesce(old.eid, new.eid)
    old.given = coalesce(old.given, new.given)
    old.access = coalesce(old.access, new.access)
    old.title = coalesce(old.title, new.title)
    old.name = coalesce(old.name, new.name)
    old.level = coalesce(old.level, new.level)
    old.external = old.external
    old.prepare = coalesce(old.prepare, new.prepare)


@commands.merge.register(Context, Manifest, Resource, NotAvailable)
def merge(context: Context, manifest: Manifest, old: Resource, new: NotAvailable) -> None:
    return


@commands.merge.register(Context, Manifest, NotAvailable, Model, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Model, has_manifest_priority: bool) -> None:
    old = copy(new)
    old.external = copy(old.external)
    if old.external and old.external.resource:
        resources = zipitems(
            old.external.dataset.resources.values(),
            [old.external.resource],
            _resource_source_key
        )
        for res in resources:
            for old_res, new_res in res:
                if old_res and new_res:
                    old.external.resource = old_res
                    old_res.models[old.name] = old
    old.manifest = manifest
    manifest.models[old.name] = old

    _merge_model_properties(context, manifest, old, new, has_manifest_priority)


@commands.merge.register(Context, Manifest, Model, Model, bool)
def merge(context: Context, manifest: Manifest, old: Model, new: Model, has_manifest_priority: bool) -> None:
    old.description = coalesce(new.description, old.description)
    old.comments = coalesce(new.comments, old.comments)
    old.lang = coalesce(new.lang, old.lang)

    old.id = coalesce(old.id, new.id)
    old.eid = coalesce(old.eid, new.eid)
    old.base = coalesce(old.base, new.base)
    old.given = coalesce(old.given, new.given)
    old.name = coalesce(old.name, new.name)
    old.title = coalesce(old.title, new.title)
    old.access = coalesce(old.access, new.access)
    old.level = coalesce(old.level, new.level)
    old.external = coalesce(old.external, new.external)
    old.manifest = manifest

    manifest.models[old.name] = old
    _merge_model_properties(context, manifest, old, new, has_manifest_priority)

    if old.external and new.external:
        if not has_manifest_priority or (old.external.unknown_primary_key and not new.external.unknown_primary_key):
            keys = zipitems(
                old.properties.values(),
                new.external.pkeys,
                _property_source_key
            )
            new_keys = []
            for key in keys:
                for o, n in key:
                    if o and n:
                        new_keys.append(o)
            if new_keys:
                old.external.pkeys = new_keys
                old.external.unknown_primary_key = new.external.unknown_primary_key


@commands.merge.register(Context, Manifest, Model, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Model, new: NotAvailable, has_manifest_priority: bool) -> None:
    if old.external and not old.external.name:
        for prop in old.properties.values():
            if prop.external:
                prop.external.name = None


@commands.merge.register(Context, Manifest, NotAvailable, Property, bool)
def merge(context: Context, manifest: Manifest, old: NotAvailable, new: Property, has_manifest_priority: bool) -> None:
    if new.external:
        new.external.prop = new
    new.dtype.prop = new
    merge(context, manifest, new.dtype, new.dtype)
    new.model.properties[new.name] = new


@commands.merge.register(Context, Manifest, Property, Property, bool)
def merge(context: Context, manifest: Manifest, old: Property, new: Property, has_manifest_priority: bool) -> None:
    if not has_manifest_priority:
        merged = old
        merged.type = coalesce(new.type, old.type)
        merged.uri = coalesce(new.uri, old.uri)
        merged.description = coalesce(new.description, old.description)
        merged.enum = coalesce(new.enum, old.enum)
        merged.enums = coalesce(new.enums, old.enums)
        merged.unit = coalesce(new.unit, old.unit)
        merged.comments = coalesce(new.comments, old.comments)
        merged.lang = coalesce(new.lang, old.lang)

        merged.model = coalesce(old.model, new.model)
        merged.external = coalesce(old.external, new.external)
        merged.place = coalesce(old.place, new.place)
        merged.name = coalesce(old.name, new.name)
        merged.given = coalesce(old.given, new.given)
        merged.title = coalesce(old.title, new.title)
        merged.level = coalesce(old.level, new.level)
        merged.access = coalesce(old.access, new.access)
        merged.dtype.prop = merged

        old.model.properties[old.name] = merged

        merge(context, manifest, merged.dtype, new.dtype)


@commands.merge.register(Context, Manifest, Property, NotAvailable, bool)
def merge(context: Context, manifest: Manifest, old: Property, new: NotAvailable, has_manifest_priority: bool) -> None:
    if old.external:
        old.external.name = None
    manifest.models[old.model.name].properties[old.name] = old


@commands.merge.register(Context, Manifest, DataType, Array)
def merge(context: Context, manifest: Manifest, old: DataType, new: Array) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)
    models = zipitems(
        [merged.items.model],
        manifest.models.values(),
        _model_source_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                properties = zipitems(
                    merged.items,
                    o.properties.values(),
                    _property_source_key
                )
                for prop in properties:
                    for po, pn in prop:
                        if po and pn:
                            merged.items = po
                            break
                    else:
                        continue
                    break
            break
        else:
            continue
        break

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Object)
def merge(context: Context, manifest: Manifest, old: DataType, new: Object) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    new_properties = {}
    for key, value in merged.properties.items():
        new_key = key
        new_value = value
        models = zipitems(
            [value.model],
            manifest.models.values(),
            _model_source_key
        )
        for model in models:
            for o, n in model:
                if o and n:
                    properties = zipitems(
                        merged.items,
                        o.properties.values(),
                        _property_source_key
                    )
                    for prop in properties:
                        for po, pn in prop:
                            if po and pn:
                                new_value = po
                                new_key = po.name
                                break
                        else:
                            continue
                        break
                break
            else:
                continue
            break
        new_properties[new_key] = new_value
    merged.properties = new_properties

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Ref)
def merge(context: Context, manifest: Manifest, old: DataType, new: Ref) -> None:
    merged = copy(new)
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    models = zipitems(
        [merged.model],
        manifest.models.values(),
        _model_source_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                if n.external and merged.prop.model.external:
                    if n.external.dataset is merged.prop.model.external.dataset:
                        merged.model = n
                        new_refprops = []
                        properties = zipitems(
                            merged.refprops,
                            n.properties.values(),
                            _property_source_key
                        )
                        for prop in properties:
                            for po, pn in prop:
                                if po and pn:
                                    new_refprops.append(pn)

                        if not new_refprops:
                            new_refprops = merged.refprops
                        merged.refprops = new_refprops
                        break
        else:
            continue
        break
    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, Denorm)
def merge(context: Context, manifest: Manifest, old: DataType, new: Denorm) -> None:
    merged = new
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)
    merged.name = coalesce(new.name, old.name)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)

    models = zipitems(
        [merged.rel_prop.model],
        manifest.models.values(),
        _model_source_key
    )
    for model in models:
        for o, n in model:
            if o and n:
                properties = zipitems(
                    merged.items,
                    o.properties.values(),
                    _property_source_key
                )
                for prop in properties:
                    for po, pn in prop:
                        if po and pn:
                            merged.rel_prop = po
                            break
                    else:
                        continue
                    break
            break
        else:
            continue
        break

    old.prop.dtype = merged


@commands.merge.register(Context, Manifest, DataType, DataType)
def merge(context: Context, manifest: Manifest, old: DataType, new: DataType) -> None:
    merged = new
    merged.name = coalesce(new.name, old.name)
    merged.type_args = coalesce(new.type_args, old.type_args)
    merged.unique = coalesce(new.unique, old.unique)
    merged.nullable = coalesce(new.nullable, old.nullable)
    merged.required = coalesce(new.required, old.required)
    merged.default = coalesce(new.default, old.default)

    merged.backend = coalesce(old.backend, new.backend)
    merged.prop = coalesce(old.prop, new.prop)
    merged.choices = coalesce(old.choices, new.choices)
    merged.prepare = coalesce(old.prepare, new.prepare)
    old.prop.dtype = merged


def _filter_models_for_dataset(
    manifest: Manifest,
    dataset: Dataset
) -> List[Model]:
    models = []
    for model in manifest.models.values():
        if model.external:
            if model.external.dataset is dataset:
                models.append(model)
    return models


def _merge_model_properties(
    context: Context,
    manifest: Manifest,
    old: Model,
    new: Model,
    has_manifest_priority: bool
):
    properties = zipitems(
        old.properties.values(),
        new.properties.values(),
        _property_source_key,
    )
    deduplicator = Deduplicator("_{}")
    for prop in properties:
        for o, n in prop:
            if n:
                n = copy(n)
            if o:
                deduplicator(o.basename)
            if n and not o:
                name = deduplicator(n.basename)
                n.model = old
                n.name = name
                n.place = name
            commands.merge(context, manifest, o, n, has_manifest_priority)


def _merge_prefixes(
    context: Context,
    manifest: Manifest,
    old: Dataset,
    new: Dataset
):
    prefixes = zipitems(
        old.prefixes.values(),
        new.prefixes.values(),
        _name_key,
    )
    for prefix in prefixes:
        for o, n in prefix:
            commands.merge(context, manifest, o, n)


def _merge_resources(
    context: Context,
    manifest: Manifest,
    old: Dataset,
    new: Dataset
):
    old_resources = [] if old == NA else old.resources.values()
    new_resources = [] if new == NA else new.resources.values()
    resources = zipitems(
        old_resources,
        new_resources,
        _resource_source_key,
    )
    for res in resources:
        for o, n in res:
            commands.merge(context, manifest, o, n)


TItem = TypeVar('TItem')


def zipitems(
    a: TItem,
    b: TItem,
    key: Callable[[TItem], Hashable],
) -> Iterator[List[Tuple[TItem, TItem]]]:

    res: Dict[
        Hashable,  # key
        List[
            Tuple[
                Any,   # a
                Any,   # b
            ]
        ]
        ,
    ] = {}

    for v in a:
        k = key(v)
        if isinstance(k, Tuple):
            for keys in res.keys():
                if set(k) == set(keys):
                    res[keys].append([v, NA])
                    break
            else:
                res[k] = [[v, NA]]
        else:
            if k not in res.keys():
                res[k] = []
            res[k].append([v, NA])
    for v in b:
        k = key(v)
        if isinstance(k, Tuple):
            found = False
            for keys in res.keys():
                if set(k).issubset(set(keys)):
                    found = True
                    additional = []
                    for item in res[keys]:
                        if item[1] is NA:
                            item[1] = v
                        else:
                            additional.append([item[0], v])
                    res[keys] += additional
            if not found:
                res[k] = [[NA, v]]
        else:
            if k in res:
                additional = []
                for item in res[k]:
                    if item[1] is NA:
                        item[1] = v
                    else:
                        additional.append([item[0], v])
                res[k] += additional
            else:
                res[k] = [[NA, v]]
    yield from res.values()


def coalesce(*args: Any) -> Any:
    for arg in args:
        if arg:
            return arg
    return arg


def _name_key(node: Node) -> str:
    return node.name


def _dataset_resource_source_key(dataset: Dataset) -> Tuple:
    keys = []
    for resource in dataset.resources.values():
        keys.append(_resource_source_key(resource))
    return tuple(keys)


def _property_source_key(prop: Property) -> str:
    if prop.external and prop.external.name:
        return prop.external.name
    else:
        return prop.name


def _resource_source_key(resource: Resource) -> str:
    result = resource.name
    manifest = resource.dataset.manifest
    if resource.external:
        result = resource.external
    else:
        if manifest.backends:
            if resource.backend.name in manifest.backends:
                result = manifest.backends[resource.backend.name].config['dsn']
        elif manifest.store.backends:
            if resource.backend.name in manifest.store.backends:
                result = manifest.store.backends[resource.backend.name].config['dsn']
    if "@" in result:
        result = result.split("@")[1]
    return result


def _backend_dsn(backend: ExternalBackend) -> str:
    dsn = backend.config['dsn']
    if "@" in dsn:
        dsn = dsn.split("@")[1]
    return dsn


def _model_source_key(model: Model) -> str:
    if model.external and model.external.name:
        return f'{_resource_source_key(model.external.resource)}/{model.external.name}'
    else:
        return model.name



from spinta.components import Context, Command, CommandList, Node
from spinta.commands import load


@load.register(Context, Command, dict)
def load(
    context: Context, command: Command, data: dict, *,
    parent: Node,
    scope: str,
    argname: str = None,
) -> Command:
    config = context.get('config')
    name, args = next(iter(data.items()))
    command.name = name
    command.parent = parent
    command.command = config.commands[scope][name]
    if isinstance(args, str):
        args = {command.command.schema.get('argname', argname): args}
    command.args = args
    return command


@load.register(Context, CommandList, list)
def load(
    context: Context, command: CommandList, data: list, *,
    parent: Node,
    scope: str,
    argname: str = None,
) -> CommandList:
    command.parent = parent
    command.commands = [
        load(context, Command(), x, parent=parent, scope=scope, argname=argname)
        for x in data
    ]
    return command



from typing import Iterator

from spinta import commands
from spinta.backends import Backend
from spinta.components import Context
from spinta.datasets.components import Dataset
from spinta.manifests.components import Manifest
from spinta.manifests.components import ManifestSchema
from spinta.manifests.helpers import dataset_to_schema
from spinta.manifests.helpers import resource_to_schema


@commands.inspect.register(Context, Backend, Manifest, type(None))
def inspect(
    context: Context,
    backend: Backend,
    manifest: Manifest,
    source: None,
) -> Iterator[ManifestSchema]:
    for dataset in manifest.datasets.values():
        yield from commands.inspect(context, manifest.backend, dataset, None)


@commands.inspect.register(Context, Backend, Dataset, type(None))
def inspect(
    context: Context,
    backend: Backend,
    dataset: Dataset,
    source: None,
) -> Iterator[ManifestSchema]:
    eid, schema = dataset_to_schema(dataset)
    schema['resources'] = {
        resource.name: resource_to_schema(resource)[1]
        for resource in dataset.resources.values()
    }
    yield eid, schema
    for resource in dataset.resources.values():
        yield from commands.inspect(context, resource.backend, resource, None)



import logging

from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest
from spinta.manifests.helpers import load_manifest_nodes
from spinta.manifests.internal.components import InternalManifest
from spinta.manifests.yaml.helpers import read_manifest_schemas
from spinta.manifests.yaml.helpers import read_freezed_manifest_schemas

log = logging.getLogger(__name__)


@commands.load.register(Context, InternalManifest)
def load(
    context: Context,
    manifest: InternalManifest,
    *,
    into: Manifest = None,
    freezed: bool = True,
    rename_duplicates: bool = False,
    load_internal: bool = True,
):
    if freezed:
        if into:
            log.info(
                'Loading freezed manifest %r into %r from %s.',
                manifest.name,
                into.name,
                manifest.path,
            )
        else:
            log.info(
                'Loading freezed manifest %r from %s.',
                manifest.name,
                manifest.path,
            )
        schemas = read_freezed_manifest_schemas(manifest)
    else:
        if into:
            log.info(
                'Loading manifest %r into %r from %s.',
                manifest.name,
                into.name,
                manifest.path,
            )
        else:
            log.info(
                'Loading manifest %r from %s.',
                manifest.name,
                manifest.path,
            )
        schemas = read_manifest_schemas(manifest)

    if into:
        load_manifest_nodes(context, into, schemas, source=manifest)
    else:
        load_manifest_nodes(context, manifest, schemas)



from spinta import commands
from spinta.components import Context
from spinta.backends.postgresql.components import PostgreSQL


@commands.bootstrap.register(Context, PostgreSQL)
def bootstrap(context: Context, backend: PostgreSQL):
    # XXX: I found, that this some times leaks connection, you can check that by
    #      comparing `backend.engine.pool.checkedin()` before and after this
    #      line.
    # TODO: update appropriate rows in _schema and save `applied` date
    #       of schema migration
    backend.schema.create_all(checkfirst=True)



import sys
from pathlib import Path
from typing import Any
from typing import Dict

from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.sqldump.components import SqlDump


@commands.load.register(Context, SqlDump, dict)
def load(context: Context, backend: SqlDump, config: Dict[str, Any]):
    if config['dsn'] == '-':
        backend.stream = sys.stdin
    elif isinstance(config['dsn'], str):
        backend.path = Path(config['dsn'])
    else:
        backend.path = config['dsn'] or None



import datetime
from decimal import Decimal
from typing import Any, Union, Tuple
from typing import Dict
from typing import List
from typing import Optional

from lxml import etree
from lxml.etree import Element, QName
from starlette.requests import Request
from starlette.responses import StreamingResponse

from spinta import commands
from spinta.backends.components import SelectTree
from spinta.backends.helpers import get_model_reserved_props
from spinta.backends.helpers import select_model_props
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.dimensions.prefix.components import UriPrefix
from spinta.formats.components import Format
from spinta.formats.html.helpers import get_model_link
from spinta.formats.rdf.components import Rdf
from spinta.types.datatype import DataType
from spinta.types.datatype import File
from spinta.types.datatype import Ref
from spinta.types.datatype import Date
from spinta.types.datatype import Time
from spinta.types.datatype import DateTime
from spinta.types.datatype import Number
from spinta.utils.schema import NotAvailable

RDF = "rdf"
PAV = "pav"
DESCRIPTION = "Description"


def _get_available_prefixes(model: Model) -> dict:
    prefixes = {}
    if model.manifest.datasets.get(model.ns.name):
        prefixes = model.manifest.datasets.get(model.ns.name).prefixes
        for key, val in prefixes.items():
            if isinstance(val, UriPrefix):
                prefixes[key] = val.uri
    return prefixes


def _get_attribute_name(
    name: str,
    prefix: str,
    available_prefixes: dict
) -> Union[str, QName]:
    if prefix and available_prefixes.get(prefix):
        attribute_name = QName(available_prefixes.get(prefix), name)
    else:
        attribute_name = name
    return attribute_name


def _get_attributes(
    model: Model,
    data: dict,
    about_name: Union[str, QName],
    type_name: Union[str, QName],
    revision_name: Union[str, QName],
) -> Tuple[dict, dict]:
    attributes = {}
    if '_id' in data:
        _id = data.pop('_id')
        if _id is not None and _id.text:
            attributes[about_name] = get_model_link(model, pk=_id.text)
    if '_type' in data:
        _type = data.pop('_type')
        if _type is not None and _type.text:
            attributes[type_name] = _type.text
    if '_revision' in data:
        _revision = data.pop('_revision')
        if _revision is not None and _revision.text:
            attributes[revision_name] = _revision.text
    return data, attributes


def _get_prefix_and_name(uri: str) -> Tuple[str, str]:
    prefix = name = None
    if uri and ':' in uri:
        prefix = uri.split(':')[0]
        name = uri.split(':')[1]
    return prefix, name


def _create_element(
    name: Union[str, QName],
    text: str = None,
    base: str = None,
    nsmap: dict = None,
    attributes: dict = None,
    children: List[Element] = None,
) -> Element:
    if attributes is None:
        attributes = {}
    if children is None:
        children = []

    if nsmap is not None:
        elem = Element(name, nsmap=nsmap)
    else:
        elem = Element(name)

    if base is not None:
        elem.base = base
    if text is not None:
        elem.text = text
    for key, value in attributes.items():
        elem.attrib[key] = value
    for child in children:
        if child is not None:
            elem.append(child)
    return elem


def _create_model_element(
    model: Model,
    data: dict
) -> Element:
    prefixes = _get_available_prefixes(model)
    about_name = _get_attribute_name('about', RDF, prefixes)
    type_name = _get_attribute_name('type', RDF, prefixes)
    revision_name = _get_attribute_name('version', PAV, prefixes)
    model_prefix, model_name = _get_prefix_and_name(model.uri)

    data, attributes = _get_attributes(
        model,
        data,
        about_name,
        type_name,
        revision_name
    )
    if model_name:
        name = _get_attribute_name(model_name, model_prefix, prefixes)
    else:
        name = _get_attribute_name(model.basename, model_prefix, prefixes)
    if not isinstance(name, QName):
        name = _get_attribute_name(DESCRIPTION, RDF, prefixes)
        attributes[type_name] = model.model_type()

    return _create_element(
        name=name,
        attributes=attributes,
        children=list(data.values()),
        nsmap=prefixes
    )


def _prepare_for_print(
    model: Model,
    data: dict
) -> str:
    elem = _create_model_element(model, data)
    res = etree.tostring(elem, encoding="unicode", pretty_print=True)
    for key, val in elem.nsmap.items():
        res = res.replace(f'xmlns:{key}="{val}" ', "")
    return res


@commands.render.register(Context, Request, Model, Rdf)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Rdf,
    *,
    action: Action,
    params: UrlParams,
    data,
    status_code: int = 200,
    headers: Optional[dict] = None,
):
    headers = headers or {}
    headers['Content-Disposition'] = f'attachment; filename="{model.basename}.rdf"'

    return StreamingResponse(
        _stream(
            request,
            model,
            action,
            data
        ),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )


async def _stream(
    request: Request,
    model: Model,
    action: Action,
    data
):
    namespaces = []
    prefixes = _get_available_prefixes(model)
    root_name = _get_attribute_name(RDF.capitalize(), RDF, prefixes)
    for key, val in prefixes.items():
        namespaces.append(f'xmlns:{key}="{val}"')
    if request.base_url:
        namespaces.append(f'xml:base="{str(request.base_url)}"')
    if isinstance(root_name, QName):
        root_name = f"{RDF}:{root_name.localname}"
    namespaces = "\n ".join(namespaces)

    yield f'<?xml version="1.0" encoding="UTF-8"?>\n<{root_name}\n {namespaces}>\n'
    if action == Action.GETONE:
        yield _prepare_for_print(model, data)
    else:
        for row in data:
            yield _prepare_for_print(model, row)
    yield f"</{root_name}>\n"


@commands.prepare_data_for_response.register(Context, Model, Rdf, dict)
def prepare_data_for_response(
    context: Context,
    model: Model,
    fmt: Rdf,
    value: dict,
    *,
    action: Action,
    select: SelectTree,
    prop_names: List[str],
) -> dict:
    value = value.copy()
    reserved = get_model_reserved_props(action)

    available_prefixes = _get_available_prefixes(model)

    value['_available_prefixes'] = available_prefixes
    value['_about_name'] = _get_attribute_name('about', RDF, available_prefixes)
    value['_resource_name'] = _get_attribute_name('resource', RDF, available_prefixes)
    value['_type_name'] = _get_attribute_name('type', RDF, available_prefixes)
    value['_revision_name'] = _get_attribute_name('version', PAV, available_prefixes)

    data = {}
    for prop, val, sel in select_model_props(
        model,
        prop_names,
        value,
        select,
        reserved,
    ):
        prefix, name = _get_prefix_and_name(prop.uri)
        if name:
            value['_elem_name'] = _get_attribute_name(name, prefix, available_prefixes)
        else:
            value['_elem_name'] = _get_attribute_name(prop.name, prefix, available_prefixes)

        elem = commands.prepare_dtype_for_response(
            context,
            fmt,
            prop.dtype,
            val,
            data=value,
            action=action,
            select=sel
        )
        data[prop.name] = elem
    return data


@commands.prepare_dtype_for_response.register(Context, Rdf, DataType, type(None))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: DataType,
    value: None,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return None


@commands.prepare_dtype_for_response.register(Context, Rdf, DataType, object)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: DataType,
    value: Any,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Date, datetime.date)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Date,
    value: datetime.date,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=value.isoformat()
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Time, datetime.time)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Time,
    value: datetime.time,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=value.isoformat()
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, DateTime, datetime.datetime)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: DateTime,
    value: datetime.datetime,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=value.isoformat()
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Number, Decimal)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: Number,
    value: Decimal,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return _create_element(
        name=data['_elem_name'],
        text=str(value)
    )


@commands.prepare_dtype_for_response.register(Context, Rdf, Ref, (dict, str, type(None)))
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: File,
    value,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    super_ = commands.prepare_dtype_for_response[Context, Format, Ref, dict]
    data_dict = super_(context, fmt, dtype, value, data=data, action=action, select=select)
    prefixes = data['_available_prefixes']
    attributes = {}
    children = []

    if len(data_dict) == 1 and '_id' in data_dict:
        if data_dict['_id'] is not None and data_dict['_id'].text:
            attributes[data['_resource_name']] = get_model_link(
                dtype.model,
                pk=data_dict['_id'].text
            )
    else:
        # TODO: write tests for denormalized form after task #216 is done

        if isinstance(data['_elem_name'], QName):
            prefix, name = _get_prefix_and_name(dtype.model.uri)
            if name:
                ref_model_name = _get_attribute_name(name, prefix, prefixes)
            else:
                ref_model_name = _get_attribute_name(dtype.model.basename, prefix, prefixes)
            data_dict, ref_model_attrs = _get_attributes(
                dtype.model,
                data_dict,
                data['_about_name'],
                data['_type_name'],
                data['_revision_name'],
            )
            ref_model_elem = _create_element(
                name=ref_model_name,
                attributes=ref_model_attrs,
                children=list(data_dict.values())
            )
            children.append(ref_model_elem)
        else:
            data_dict['_type'] = _create_element(
                name='_type',
                text=dtype.model.model_type()
            )
            data_dict, description_attrs = _get_attributes(
                dtype.model,
                data_dict,
                data['_about_name'],
                data['_type_name'],
                data['_revision_name'],
            )
            description_elem = _create_element(
                name=_get_attribute_name(DESCRIPTION, RDF, prefixes),
                attributes=description_attrs,
                children=list(data_dict.values())
            )
            children.append(description_elem)

    if attributes or children:
        return _create_element(
            name=data['_elem_name'],
            attributes=attributes,
            children=children
        )
    else:
        return None


@commands.prepare_dtype_for_response.register(Context, Rdf, File, NotAvailable)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: File,
    value: NotAvailable,
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    return None


@commands.prepare_dtype_for_response.register(Context, Rdf, File, dict)
def prepare_dtype_for_response(
    context: Context,
    fmt: Rdf,
    dtype: File,
    value: Dict[str, Any],
    *,
    data: Dict[str, Any],
    action: Action,
    select: dict = None
):
    if value['_id'] is not None:
        attributes = {
            data['_about_name']: get_model_link(
                dtype.prop.model,
                pk=data['_id'],
                prop=dtype.prop.name
            )
        }
        return _create_element(
            name=data['_elem_name'],
            attributes=attributes
        )
    else:
        return None



from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.ufuncs.components import Engine
from spinta.datasets.backends.sql.ufuncs.components import SqlResource
from spinta.exceptions import UnknownBind


@ufunc.resolver(SqlResource, Bind)
def connect(env: SqlResource, bind: Bind, **kwargs) -> Engine:
    if bind.name == 'self':
        return env.call('connect', env.dsn, **kwargs)
    else:
        raise UnknownBind(name=bind.name)


@ufunc.resolver(SqlResource, str)
def connect(env: SqlResource, dsn: str, **kwargs) -> Engine:
    return Engine(dsn, **kwargs)



from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Set

import itertools


def flatten(value, sep='.'):
    value, lists = _flatten(value, sep)

    if value is None:
        for k, vals in lists:
            for v in vals:
                yield from flatten(v, sep)

    elif lists:
        keys, lists = zip(*lists)
        for vals in itertools.product(*lists):
            val = {
                sep.join(k): v
                for k, v in zip(keys, vals)
            }
            val.update(value)
            yield from flatten(val, sep)

    else:
        yield value


def _flatten(value, sep, key=()):
    if isinstance(value, dict):
        data = {}
        lists = []
        for k, v in value.items():
            v, more = _flatten(v, sep, key + (k,))
            data.update(v or {})
            lists += more
        return data, lists

    elif isinstance(value, (list, Iterator)):
        if value:
            return None, [(key, value)]
        else:
            return None, []

    else:
        return {sep.join(key): value}, []


def build_select_tree(select: List[str]) -> Dict[str, Set[Optional[str]]]:
    tree: Dict[str, Set[Optional[str]]] = {}
    for name in select:
        split = name, None
        while len(split) == 2:
            name, node = split
            if name not in tree:
                tree[name] = set()
            if node:
                tree[name].add(node)
            split = name.rsplit('.', 1)
    return tree


def flat_dicts_to_nested(value):
    res = {}
    for k, v in dict(value).items():
        names = k.split('.')
        vref = res
        for name in names[:-1]:
            if name not in vref:
                vref[name] = {}
            vref = vref[name]
        vref[names[-1]] = v
    return res


def flatten_value(value, sep=".", key=()):
    value, _ = _flatten(value, sep, key)
    return value



from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Denorm
from spinta.backends.postgresql.components import PostgreSQL


@commands.prepare.register(Context, PostgreSQL, Denorm)
def prepare(context: Context, backend: PostgreSQL, dtype: Denorm):
    rel_prop = dtype.rel_prop
    original_place = rel_prop.place
    rel_prop.place = dtype.prop.place
    column = commands.prepare(
        context,
        backend,
        rel_prop.dtype,
    )
    rel_prop.place = original_place
    return column



from __future__ import annotations

import dataclasses
from builtins import staticmethod
from typing import Any

from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import TypedDict
from typing import Union

from spinta.components import Component
from spinta.components import Mode
from spinta.components import Model
from spinta.components import Namespace
from spinta.components import Store
from spinta.core.enums import Access
from spinta.dimensions.enum.components import Enums
from spinta.dimensions.prefix.components import UriPrefix

if TYPE_CHECKING:
    from spinta.backends.components import Backend
    from spinta.datasets.components import Dataset
    from spinta.datasets.keymaps.components import KeyMap


class MetaDataContainer(TypedDict):
    ns: Dict[str, Namespace]
    dataset: Dict[str, Dataset]
    model: Dict[str, Model]


class ManifestGiven:
    access: str = None


class Manifest(Component):
    type: str = None
    name: str = None
    keymap: KeyMap = None
    backend: Backend = None
    parent: Component = None
    store: Store = None
    objects: MetaDataContainer = None
    path: str = None
    access: Access = Access.protected
    prefixes: Dict[str, UriPrefix]
    enums: Enums

    # {<endpoint>: <model.name>} mapping. There can be multiple model types,
    # but name and endpoint for all of them should match.
    endpoints: Dict[str, str] = None

    # Backends defined in the manifest.
    backends: Dict[str, Backend] = None

    # List of other source manifests used to populate nodes into the main
    # manifest.
    sync: List[Manifest]

    mode: Mode = Mode.internal

    given: ManifestGiven

    @staticmethod
    def detect_from_path(path: str) -> bool:
        return False

    def __init__(self):
        self.given = ManifestGiven()

    def __repr__(self):
        return (
            f'<{type(self).__module__}.{type(self).__name__}'
            f'(name={self.name!r})>'
        )

    def __eq__(self, other: Union[Manifest, str]):
        if isinstance(other, str):
            # A hack for tests, to make possible things like:
            #     assert manifest == """d | r | m | ..."""
            # This uses pytest_assertrepr_compare hook and compare_manifest to
            # eventually compare manifests in ascii table form.
            from spinta.testing.manifest import compare_manifest
            left, right = compare_manifest(self, other)
            return left == right
        else:
            super().__eq__(other)

    def add_model_endpoint(self, model):
        endpoint = model.endpoint
        if endpoint:
            if endpoint not in self.endpoints:
                self.endpoints[endpoint] = model.name
            elif self.endpoints[endpoint] != model.name:
                raise Exception(
                    f"Same endpoint, but different model name, "
                    f"endpoint={endpoint!r}, model.name={model.name!r}."
                )

    @property
    def models(self) -> Dict[str, Model]:
        return self.objects['model']

    @property
    def datasets(self) -> Dict[str, Dataset]:
        return self.objects['dataset']

    @property
    def namespaces(self) -> Dict[str, Namespace]:
        return self.objects['ns']


NodeSchema = Optional[Dict[str, Any]]
ManifestSchema = Tuple[Any, NodeSchema]


@dataclasses.dataclass
class ManifestPath:
    type: str = 'tabular'
    name: str = None
    path: str = None
    file: IO = None



from typing import Callable
from typing import Union

from typer import Typer


def add(app: Typer, name: str, item: Union[Callable, Typer], **kwargs):
    if isinstance(item, Typer):
        app.add_typer(item, name=name, **kwargs)
    else:
        app.command(name, **kwargs)(item)



from spinta import commands
from spinta.components import Context
from spinta.types.datatype import Array
from spinta.types.helpers import set_dtype_backend


@commands.link.register(Context, Array)
def link(context: Context, dtype: Array) -> None:
    set_dtype_backend(dtype)
    # In case of a dynamic array, dtype of items is not known.
    if dtype.items:
        commands.link(context, dtype.items.dtype)



from typing import List, Iterable

from pathlib import Path

import jsonpatch
import requests
import httpx

from ruamel.yaml import YAML

from spinta.utils.schema import NA

yaml = YAML(typ='safe')


def create_manifest_files(tmp_path, manifest):
    for file, data in manifest.items():
        if data is None:
            data = Path('tests/manifest') / file
            data = next(yaml.load_all(data.read_text()))
        path = tmp_path / file
        path.parent.mkdir(parents=True, exist_ok=True)
        if isinstance(data, dict):
            yaml.dump(data, path)
        elif isinstance(data, list):
            yaml.dump_all(data, path)


def update_manifest_files(tmp_path, manifest):
    for file, patch in manifest.items():
        path = tmp_path / file
        versions = list(yaml.load_all(path.read_text()))
        patch = jsonpatch.JsonPatch(patch)
        versions[0] = patch.apply(versions[0])
        with path.open('w') as f:
            yaml.dump_all(versions, f)


def read_manifest_files(tmp_path):
    path = tmp_path
    yaml = YAML(typ='safe')
    manifests = {}
    for fp in path.glob('**/*.yml'):
        data = list(yaml.load_all(fp.read_text()))
        manifest_file = str(fp)[len(str(tmp_path)) + 1:]
        manifests[manifest_file] = data
    return manifests


def readable_manifest_files(manifest):
    ids = {}
    for filename, versions in manifest.items():
        manifest[filename] = [versions[0]]
        version = versions[0]
        name = version['name']

        # Fill ids dict
        for i, version in enumerate(versions):
            if 'id' in version:
                ids[version['id']] = f'{name}:{i}'

        # Read only last version.
        for i, version in enumerate(versions[-1:], len(versions) - 1):
            del version['date']
            del version['changes']
            for action in version['migrate']:
                action['upgrade'] = action['upgrade'].splitlines()
                action['downgrade'] = action['downgrade'].splitlines()
            manifest[filename] += [version]

    # Replaces UUIDs to human version numbers
    for filename, versions in manifest.items():
        for version in versions:
            if 'id' in version:
                version['id'] = ids.get(version['id'], '(UNKNOWN)')
            if 'version' in version and isinstance(version['version'], str):
                version['version'] = ids.get(version['version'], '(UNKNOWN)')

    return manifest


def errors(resp, *keys, status: int = None):
    errors = _get_errors_data(resp, status)
    return [_extract_error(err, keys) for err in errors]


def error(resp, *keys, status: int = None):
    errors = _get_errors_data(resp, status)
    assert len(errors) == 1, errors
    return _extract_error(errors[0], keys)


def _get_errors_data(resp, status: int = None):
    if resp.headers['content-type'].startswith('text/html'):
        data = resp.context
    else:
        data = resp.json()
    status = status or resp.status_code
    assert resp.status_code == status, data
    assert 'errors' in data, data
    return data['errors']


def _extract_error(err: dict, keys: Iterable[str]):
    res = {}
    if keys:
        for k in keys:
            if isinstance(k, list):
                ctx = err.get('context', {})
                res['context'] = {k: ctx.get(k, NA) for k in k}
            else:
                res[k] = err.get(k, NA)
    else:
        return err.get('code', NA)
    return res


def get_error_codes(response):
    assert 'errors' in response, response
    return [err['code'] for err in response['errors']]


def get_error_context(response, error_code, ctx_keys: List[str] = None):
    assert response["errors"]
    for err in response["errors"]:
        # FIXME: probably shouldn't return the context of the first found error
        # If more than one error is found and it's not intended - error
        # should be thrown
        if err["code"] == error_code:
            if ctx_keys is None:
                return err["context"]
            else:
                return {k: v for k, v in err["context"].items() if k in ctx_keys}
    else:
        assert False


class RowIds:

    def __init__(self, ids, id_key='_id', search_resp_key='_data'):
        self.id_key = id_key
        self.search_resp_key = search_resp_key
        self.ids = {k: v for v, k in enumerate(self._cast(ids))}
        self.idsr = {k: v for v, k in self.ids.items()}

    def __call__(self, ids):
        return [self.ids.get(i, i) for i in self._cast(ids)]

    def __getitem__(self, i):
        return self.idsr[i]

    def _cast(self, ids):
        if isinstance(ids, (requests.models.Response, httpx.Response)):
            resp = ids
            assert resp.status_code == 200, resp.json()
            ids = resp.json()
        if isinstance(ids, dict):
            ids = ids[self.search_resp_key]
        if isinstance(ids, list) and len(ids) > 0 and isinstance(ids[0], dict):
            ids = [r[self.id_key] for r in ids]
        return ids



import csv
import pathlib
from io import StringIO
from typing import List

from spinta.manifests.tabular.components import ManifestColumn
from spinta.manifests.tabular.constants import DATASET
from spinta.manifests.tabular.helpers import read_ascii_tabular_rows
from spinta.manifests.tabular.helpers import torow
from spinta.manifests.tabular.helpers import write_tabular_manifest


def create_tabular_manifest(
    path: pathlib.Path,
    manifest: str,
) -> None:
    path = pathlib.Path(path)
    rows = read_ascii_tabular_rows(manifest, strip=True)
    cols: List[ManifestColumn] = next(rows, [])
    if cols:
        rows = (torow(DATASET, dict(zip(cols, row))) for row in rows)
        write_tabular_manifest(str(path), rows)


def convert_ascii_manifest_to_csv(manifest: str) -> bytes:
    file = StringIO()
    writer = csv.writer(file)
    rows = read_ascii_tabular_rows(
        manifest,
        strip=True,
        check_column_names=False,
    )
    for row in rows:
        writer.writerow(row)
    return file.getvalue().encode()



import operator
import itertools

from spinta.components import Context, Action, UrlParams, Model
from spinta.formats.ascii.helpers import get_widths, get_displayed_cols, draw_border, draw_header, draw_row
from spinta.manifests.components import Manifest
from spinta.formats.components import Format
from spinta.utils.nestedstruct import flatten
from spinta.formats.helpers import get_model_tabular_header


class Ascii(Format):
    content_type = 'text/plain'
    accept_types = {
        'text/plain',
    }
    params = {
        'width': {'type': 'integer'},
        'colwidth': {'type': 'integer'},
    }

    def __call__(
        self,
        context: Context,
        model: Model,
        action: Action,
        params: UrlParams,
        data,
        width=None,
        max_col_width=None,
        max_value_length=100,
        rows_to_check=200,
        separator="  ",
    ):
        manifest: Manifest = context.get('store').manifest

        if action == Action.GETONE:
            data = [data]

        data = iter(data)
        peek = next(data, None)

        if peek is None:
            return

        data = itertools.chain([peek], data)

        if '_type' in peek:
            groups = itertools.groupby(data, operator.itemgetter('_type'))
        else:
            groups = [(None, data)]

        for name, group in groups:
            if name:
                yield f'\n\nTable: {name}\n'
                model = manifest.models[name]

            rows = flatten(group)
            cols = get_model_tabular_header(context, model, action, params)
            read_rows, widths = get_widths(
                rows,
                cols,
                max_value_length,
                max_col_width,
                rows_to_check
            )
            rows = itertools.chain(read_rows, rows)
            if width:
                shortened, displayed_cols = get_displayed_cols(widths, width, separator)
            else:
                shortened, displayed_cols = False, cols

            yield draw_border(widths, displayed_cols, separator, shortened)
            yield draw_header(widths, displayed_cols, separator, shortened)

            for row in rows:
                yield draw_row(
                    row,
                    widths,
                    displayed_cols,
                    max_value_length,
                    separator,
                    shortened
                )

            yield draw_border(widths, displayed_cols, separator, shortened)



from typing import Any
from typing import Dict
from typing import Optional
from typing import Iterable

from starlette.requests import Request

from spinta.typing import ObjectData
from spinta import commands
from spinta.components import Context
from spinta.components import UrlParams
from spinta.components import Version
from spinta.components import Model
from spinta.components import Action
from spinta.manifests.components import Manifest
from spinta.auth import AdminToken
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.renderer import render
from spinta.formats.html.commands import build_template_context
from spinta.formats.html.components import Cell


def make_get_request(
    path: str,
    query: Optional[str] = None,
    headers: Optional[Dict[str, str]] = None,
    method: str = 'GET',
) -> Request:
    if not path.startswith('/'):
        path = '/' + path
    req = {
        'type': 'http',
        'method': method,
        'path': path,
        'path_params': {'path': path},
        'headers': [
            (k.lower().encode(), v.encode())
            for k, v in (headers or {}).items()
        ],
    }
    if query:
        req['query_string'] = query.encode()
    return Request(req)


def render_data(
    context: Context,
    manifest: Manifest,
    path: str,
    query: Optional[str],
    data: ObjectData,
    *,
    method: str = 'GET',
    accept: str = 'application/json',
    headers: Optional[Dict[str, str]] = None,
) -> Optional[Dict[str, Any]]:
    context.set('auth.token', AdminToken())

    if headers is None:
        headers = {}
    headers['Accept'] = accept
    request = make_get_request(path, query, headers, method)
    params: UrlParams = commands.prepare(
        context,
        UrlParams(),
        Version(),
        request,
    )
    action = params.action
    model = params.model

    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
        reserved=['_type', '_id', '_revision'],
    )
    data = commands.prepare_data_for_response(
        context,
        model,
        params.fmt,
        data,
        action=action,
        select=select_tree,
        prop_names=prop_names,
    )

    if params.action in (Action.GETALL, Action.SEARCH):
        data = [data]

    if params.format == 'html':
        resp = _get_html_template_context(context, model, action, params, data)
    else:
        resp = render(context, request, model, params, data, action=action)

    return resp


def _get_html_template_context(
    context: Context,
    model: Model,
    action: Action,
    params: UrlParams,
    rows: Iterable[Dict[str, Cell]],
):
    ctx = build_template_context(context, model, action, params, rows)
    resp = next(ctx['data'], None)
    if resp is not None:
        resp = dict(zip(ctx['header'], resp))
    return resp



from dataclasses import dataclass
from pathlib import Path

from spinta.core.ufuncs import Env


class PrepareFileResource(Env):
    path: Path

    def init(self, path: Path):
        return self(path=path)


@dataclass
class File:
    path: Path = None
    encoding: str = 'utf-8'

    def open(self):
        return open(self.path, 'r', encoding=self.encoding)



from spinta import commands
from spinta.components import Context
from spinta.core.config import RawConfig
from spinta.manifests.memory.components import MemoryManifest


@commands.configure.register(Context, MemoryManifest)
def configure(context: Context, manifest: MemoryManifest):
     rc: RawConfig = context.get('rc')
     manifest.path = rc.get('manifests', manifest.name, 'path')



import pathlib

from spinta import commands
from spinta.components import Context
from spinta.types.datatype import File
from spinta.formats.components import Format
from spinta.backends.fs.components import FileSystem


@commands.decode.register(Context, Format, FileSystem, File, dict)
def decode(context: Context, source: Format, target: FileSystem, dtype: File, value: dict):
    if value.get('_id'):
        value['_id'] = pathlib.Path(value['_id'])
    return value



import logging
import pathlib
import types
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import tqdm

from spinta import commands
from spinta import components
from spinta.cli.helpers.errors import ErrorCounter
from spinta.formats.components import Format
from spinta.components import Context
from spinta.components import DataStream
from spinta.components import Model
from spinta.core.ufuncs import Expr
from spinta.utils.aiotools import alist
from spinta.utils.itertools import peek

ModelRow = Tuple[Model, Dict[str, Any]]

log = logging.getLogger(__name__)


def _get_row_count(
    context: components.Context,
    model: components.Model,
) -> int:
    query = Expr('select', Expr('count'))
    stream = commands.getall(context, model, model.backend, query=query)
    for data in stream:
        return data['count()']


def count_rows(
    context: Context,
    models: List[Model],
    limit: int = None,
    *,
    stop_on_error: bool = False,
    error_counter: ErrorCounter = None
) -> Dict[str, int]:
    counts = {}
    for model in tqdm.tqdm(models, 'Count rows', ascii=True, leave=False):
        try:
            count = _get_row_count(context, model)
        except Exception:
            if error_counter:
                error_counter.increase()
                if error_counter.has_reached_max():
                    break
            if stop_on_error:
                break
            else:
                log.exception("Error on _get_row_count({model.name}).")
        else:
            if limit:
                count = min(count, limit)
            counts[model.name] = count
    return counts


def _read_model_data(
    context: components.Context,
    model: components.Model,
    limit: int = None,
    stop_on_error: bool = False,
) -> Iterable[Dict[str, Any]]:

    if limit is None:
        query = None
    else:
        query = Expr('limit', limit)

    stream = commands.getall(context, model, model.backend, query=query)

    if stop_on_error:
        stream = peek(stream)
    else:
        try:
            stream = peek(stream)
        except Exception:
            log.exception(f"Error when reading data from model {model.name}")
            return

    yield from stream


def iter_model_rows(
    context: Context,
    models: List[Model],
    counts: Dict[str, int],
    limit: int = None,
    *,
    stop_on_error: bool = False,
    no_progress_bar: bool = False,
) -> Iterator[ModelRow]:
    for model in models:
        rows = _read_model_data(context, model, limit, stop_on_error)
        if not no_progress_bar:
            count = counts.get(model.name)
            rows = tqdm.tqdm(rows, model.name, ascii=True, total=count, leave=False)
        for row in rows:
            yield model, row


async def process_stream(
    source: str,
    stream: DataStream,
    exporter: Optional[Format] = None,
    path: Optional[pathlib.Path] = None,
) -> None:
    if exporter:
        # TODO: Probably exporters should support async generators.
        if isinstance(stream, types.AsyncGeneratorType):
            stream = await alist(stream)
        chunks = exporter(stream)
        if path is None:
            for chunk in chunks:
                print(chunk, end='')
        else:
            with path.open('wb') as f:
                for chunk in chunks:
                    f.write(chunk)
    else:
        with tqdm.tqdm(desc=source, ascii=True) as pbar:
            async for _ in stream:
                pbar.update(1)


def ensure_data_dir(
    path: pathlib.Path,
):
    """Create default data directory if it does not exits"""
    path.mkdir(parents=True, exist_ok=True)



import contextlib

import sqlalchemy as sa
from sqlalchemy.engine.base import Engine

from spinta.components import Model
from spinta.components import Property
from spinta.datasets.components import ExternalBackend
from spinta.exceptions import NoExternalName
from spinta.exceptions import PropertyNotFound


class Sql(ExternalBackend):
    type: str = 'sql'
    engine: Engine = None
    schema: sa.MetaData = None
    dbschema: str = None  # Database schema name

    @contextlib.contextmanager
    def transaction(self, write=False):
        raise NotImplementedError

    @contextlib.contextmanager
    def begin(self):
        with self.engine.begin() as conn:
            yield conn

    def get_table(self, model: Model, name: str = None) -> sa.Table:
        name = name or model.external.name

        if self.dbschema:
            key = f'{self.dbschema}.{name}'
        else:
            key = name

        if key not in self.schema.tables:
            sa.Table(name, self.schema, autoload_with=self.engine)

        return self.schema.tables[key]

    def get_column(
        self,
        table: sa.Table,
        prop: Property,
        *,
        select=False,
    ) -> sa.Column:
        if prop.external is None or not prop.external.name:
            raise NoExternalName(prop)
        if prop.external.name not in table.c:
            raise PropertyNotFound(
                prop.model,
                property=prop.name,
                external=prop.external.name,
            )
        return table.c[prop.external.name]



import contextlib
import json
import os
import tempfile
from typing import Dict
from typing import List

import sqlalchemy as sa

from spinta.testing.cli import SpintaCliRunner

Schema = Dict[str, List[sa.Column]]


def pull(cli: SpintaCliRunner, rc, dataset, model=None, *, push=True):
    result = cli.invoke(rc, [
        'pull', dataset,
        ['--push'] if push else [],
        ['--model', model] if model else [],
        '-e', 'stdout:jsonl',
    ])
    data = []
    for line in result.stdout.splitlines():
        try:
            d = json.loads(line)
        except json.JSONDecodeError:
            print(line)
        else:
            data.append(d)
    return data


class Sqlite:

    def __init__(self, dsn: str):
        self.dsn = dsn
        self.engine = sa.create_engine(dsn)
        self.schema = sa.MetaData(self.engine)
        self.tables = {}

    def init(self, tables: Schema):
        self.tables = {
            k: sa.Table(k, self.schema, *v)
            for k, v in tables.items()
        }
        self.schema.create_all()

    def write(self, table, data):
        table = self.tables[table]
        with self.engine.begin() as conn:
            conn.execute(table.insert(), data)


@contextlib.contextmanager
def create_sqlite_db(tables: Schema):
    with tempfile.TemporaryDirectory() as tmpdir:
        db = Sqlite('sqlite:///' + os.path.join(tmpdir, 'db.sqlite'))
        db.init(tables)
        yield db



from spinta import spyna


def binds_to_strs(ast):
    # XXX: After replacing pyrql with spinta.spyna, AST has changed and now
    #      binds are explicit and looks like this:
    #
    #           {'name': 'bind', 'args': ['name']}
    #
    #      Previously this lookes like this:
    #
    #           'name'
    #
    #      This functions goes though all ast and converts explicit  binds to
    #      strings to make AST backwards compatible.
    #
    #      But this is a temporary solution, binds must be explicit and we need
    #      to know when user passed a bind and when a literal string.

    # Return all literals as is.
    if not isinstance(ast, dict):
        return ast

    if ast['name'] in ('bind', 'getattr'):
        return spyna.unparse(ast)

    # Process all AST recursively.
    return {
        **ast,
        'args': [binds_to_strs(x) for x in ast['args']],
    }



from typing import Iterator

import pathlib

from spinta import commands
from spinta.utils.itertools import last
from spinta.components import Context
from spinta.manifests.yaml.components import YamlManifest
from spinta.manifests.yaml.helpers import list_yaml_files
from spinta.manifests.yaml.helpers import read_yaml_file
from spinta.manifests.yaml.helpers import read_schema_versions


@commands.manifest_list_schemas.register(Context, YamlManifest)
def manifest_list_schemas(
    context: Context,
    manifest: YamlManifest,
) -> Iterator[pathlib.Path]:
    yield from list_yaml_files(manifest)


@commands.manifest_read_current.register(Context, YamlManifest)
def manifest_read_current(
    context: Context,
    manifest: YamlManifest,
    *,
    eid: pathlib.Path,
) -> dict:
    return next(read_yaml_file(eid), None)


@commands.manifest_read_freezed.register(Context, YamlManifest)
def manifest_read_freezed(
    context: Context,
    manifest: YamlManifest,
    *,
    eid: pathlib.Path,
) -> dict:
    return last(read_schema_versions(eid), default={'schema': None})['schema']


@commands.manifest_read_versions.register(Context, YamlManifest)
def manifest_read_versions(
    context: Context,
    manifest: YamlManifest,
    *,
    eid: pathlib.Path,
) -> Iterator[dict]:
    yield from read_schema_versions(eid)



import datetime
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import TypedDict

from spinta.components import Config
from spinta.components import Context
from spinta.components import Model
from spinta.components import Node
from spinta.components import Property
from spinta.components import UrlParams
from spinta.formats.html.components import Cell
from spinta.formats.html.components import Color
from spinta.manifests.components import Manifest
from spinta.types.datatype import File
from spinta.types.datatype import Ref
from spinta.utils.url import build_url_path

CurrentLocation = List[Tuple[
    str,            # Link title
    Optional[str],  # Link URL
]]


def get_current_location(
    config: Config,
    model: Model,
    params: UrlParams,
) -> CurrentLocation:
    # Remove config root
    path = params.path
    if config.root:
        if path.startswith(config.root):
            path = path[len(config.root) + 1:]
        elif config.root.startswith(path):
            path = ''

    parts = _split_path(model.manifest, config.root, path)
    if len(parts) > 0:
        parts, last = parts[:-1], parts[-1]
    else:
        parts, last = [], None

    loc: CurrentLocation = [('🏠', '/')]
    loc += [(p.title, p.link) for p in parts]

    if model.name == '_ns':
        if last is not None:
            loc += [(last.title, None)]

    else:
        pk = params.pk
        changes = params.changes

        if last is not None:
            if pk or changes:
                loc += [(last.title, get_model_link(model))]
            else:
                loc += [(last.title, None)]

        if pk:
            pks = params.pk[:8]  # short version of primary key
            if changes:
                loc += [(pks, get_model_link(model, pk=pk))]
            else:
                loc += [(pks, None)]

        if changes:
            loc += [('Changes', None)]
        else:
            loc += [('Changes', get_model_link(model, pk=pk, changes=[-10]))]

    return loc


def short_id(value: Optional[str]) -> Optional[str]:
    if value is not None:
        return value[:8]


def get_cell(
    context: Context,
    prop: Property,
    pk: Optional[str],
    row: Dict[str, Any],
    name: Any,
    *,
    shorten=False,
    color: Optional[Color] = None,
) -> Cell:
    link = None
    model = None
    if prop.dtype.name == 'ref':
        value = row.get(f'{name}._id')
    elif isinstance(prop.dtype, File):
        # XXX: In listing, row is flattened, in single object view row is
        #      nested, because of that, we need to check both cases here.
        value = row.get(f'{name}._id') or row.get(name, {}).get('_id')
        if pk:
            # Primary key might not be given in select, for example
            # select(count()).
            link = '/' + build_url_path(
                get_model_link_params(prop.model, pk=pk, prop=prop.place)
            )
    else:
        value = row.get(name)

    if prop.name == '_id' and value:
        model = prop.model
    elif isinstance(prop.dtype, Ref) and prop.dtype.model and value:
        model = prop.dtype.model

    if model:
        link = '/' + build_url_path(get_model_link_params(model, pk=value))

    if prop.dtype.name in ('ref', 'pk') and shorten and isinstance(value, str):
        value = short_id(value)

    if isinstance(value, datetime.datetime):
        value = value.isoformat()

    max_column_length = 200
    if shorten and isinstance(value, str) and len(value) > max_column_length:
        value = value[:max_column_length] + '...'

    if value is None:
        value = ''
        if color is None:
            color = Color.null

    return Cell(value, link, color)


def get_ns_data(rows) -> Iterator[List[Cell]]:
    yield [
        Cell('title'),
        Cell('description'),
    ]
    for row in rows:
        if row['title']:
            title = row['title']
        else:
            parts = row['_id'].split('/')
            if row['_type'] == 'ns':
                title = parts[-2]
            else:
                title = parts[-1]

        if row['_type'] == 'ns':
            icon = '📁'
            suffix = '/'
        else:
            icon = '📄'
            suffix = ''

        yield [
            Cell(f'{icon} {title}{suffix}', link='/' + row['_id']),
            Cell(row['description']),
        ]


class _ParsedNode(TypedDict):
    name: str
    args: List[Any]


def get_model_link_params(
    model: Node,
    *,
    pk: Optional[str] = None,
    prop: Optional[str] = None,
    **extra,
) -> List[_ParsedNode]:
    assert prop is None or (prop and pk), (
        "If prop is given, pk must be given too."
    )

    ptree = [
        {
            'name': 'path',
            'args': (
                model.name.split('/') +
                ([pk] if pk is not None else []) +
                ([prop] if prop is not None else [])
            ),
        }
    ]

    for k, v in extra.items():
        ptree.append({
            'name': k,
            'args': v,
        })

    return ptree


def get_model_link(*args, **kwargs):
    return '/' + build_url_path(get_model_link_params(*args, **kwargs))


class PathInfo(NamedTuple):
    path: str = ''
    name: str = ''
    link: str = ''
    title: str = ''


def _split_path(
    manifest: Manifest,
    base: str,
    orig_path: str,
) -> List[PathInfo]:
    parts = orig_path.split('/') if orig_path else []
    result: List[PathInfo] = []
    last = len(parts)
    base = [base] if base else []
    for i, part in enumerate(parts, 1):
        path = '/'.join(base + parts[:i])
        if i == last and path in manifest.models:
            title = manifest.models[path].title
        elif path in manifest.namespaces:
            title = manifest.namespaces[path].title
        else:
            title = ''
        title = title or part
        result.append(PathInfo(
            path=path,
            name=part,
            link=f'/{path}',
            title=title,
        ))
    return result


def get_template_context(context: Context, model, params: UrlParams):
    config: Config = context.get('config')
    return {
        'location': get_current_location(config, model, params),
    }


def _get_output_format_link(fmt: str, params: UrlParams) -> str:
    query = ''
    change = {'format': [fmt]}

    if params.changes:
        params.changes_offset = -10
        query = 'limit(100)'

    params = params.changed_parsetree(change)
    link = '/' + build_url_path(params)
    if query:
        link = link + '?' + query
    return link


def get_output_formats(params: UrlParams):
    links = [
        ('CSV', 'csv'),
        ('JSON', 'json'),
        ('JSONL', 'jsonl'),
        ('ASCII', 'ascii'),
        ('RDF', 'rdf'),
    ]
    return [
        (label, _get_output_format_link(fmt, params))
        for label, fmt in links
    ]



from spinta import commands
from spinta.components import Context, Property, Model
from spinta.exceptions import NoRefPropertyForDenormProperty, ReferencedPropertyNotFound
from spinta.types.datatype import Denorm, Ref, Object
from spinta.types.helpers import set_dtype_backend


@commands.link.register(Context, Denorm)
def link(context: Context, dtype: Denorm) -> None:
    set_dtype_backend(dtype)

    dtype.rel_prop = _get_denorm_prop(
        dtype.prop.name,
        dtype.prop,
        dtype.prop.model
    )


def _get_denorm_prop(
    name: str,
    prop: Property,
    model: Model,
) -> Property:
    models = model.manifest.models
    name_parts = name.split('.', 1)
    name = name_parts[0]
    properties = model.properties
    if len(name_parts) > 1:
        if name not in properties or not isinstance(properties[name].dtype, (Ref, Object)):
            if prop.model == model:
                raise NoRefPropertyForDenormProperty(
                    prop,
                    ref=name,
                    prop=prop.name,
                )
            else:
                raise ReferencedPropertyNotFound(
                    prop,
                    ref={'property': name, 'model': model.name},
                )
        else:
            ref_prop = properties[name]
            model = models[ref_prop.dtype.model.name]
            denorm_prop = _get_denorm_prop(name_parts[1], prop, model)
    else:
        if name not in properties:
            raise ReferencedPropertyNotFound(
                prop,
                ref={'property': name, 'model': model.name},
            )
        denorm_prop = properties[name]
    return denorm_prop



from spinta import commands
from spinta.components import Context
from spinta.manifests.yaml.components import YamlManifest


@commands.migrate.register(Context, YamlManifest)
def migrate(context: Context, manifest: YamlManifest):
    raise Exception(
        "Can't run migrations on 'yaml' manifest, use `spinta bootstrap` "
        "command instead."
    )



from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.frictionless.components import FrictionlessBackend


@commands.wait.register(Context, FrictionlessBackend)
def wait(context: Context, backend: FrictionlessBackend, *, fail: bool = False) -> bool:
    return True



from typing import Dict

from spinta import commands
from spinta.components import Config
from spinta.components import Node
from spinta.manifests.components import Manifest


@commands.get_error_context.register(Config)
def get_error_context(config: Config, *, prefix='this') -> Dict[str, str]:
    return {
        'config': f'{prefix}.rc.get_source_names()',
    }


@commands.get_error_context.register(Manifest)
def get_error_context(manifest: Manifest, *, prefix='this') -> Dict[str, str]:
    return {
        'schema': f'{prefix}.path.__str__()',
        'manifest': f'{prefix}.name'
    }


@commands.get_error_context.register(type(None))
def get_error_context(node: type(None), *, prefix='this') -> Dict[str, str]:
    return {}


@commands.get_error_context.register(Node)
def get_error_context(node: Node, *, prefix='this') -> Dict[str, str]:
    depth = 0
    context = {}
    while node:
        name = prefix + ('.' if depth else '') + '.'.join(['parent'] * depth)
        if isinstance(node, Manifest):
            context['manifest'] = f'{name}.name'
            break
        elif node.type:
            context[node.type] = f'{name}.name'
        depth += 1
        node = node.parent
    return context



from typing import Optional
from typing import overload
from pathlib import Path

from spinta.typing import ObjectData
from spinta.typing import FileObjectData
from spinta import commands
from spinta.components import Context
from spinta.components import Property
from spinta.backends.fs.components import FileSystem
from spinta.types.datatype import DataType, File
from spinta.exceptions import ItemDoesNotExist


@overload
@commands.getone.register(Context, Property, DataType, FileSystem)
def getone(
    context: Context,
    prop: Property,
    dtype: DataType,
    backend: FileSystem,
    *,
    id_: str,
) -> ObjectData:
    raise NotImplementedError


@overload
@commands.getone.register(Context, Property, File, FileSystem)
def getone(
    context: Context,
    prop: Property,
    dtype: File,
    backend: FileSystem,
    *,
    id_: str,
) -> FileObjectData:
    data = commands.getone(context, prop, dtype, prop.model.backend, id_=id_)
    if data is None:
        raise ItemDoesNotExist(prop, id=id_)
    data = (dtype.backend.path / data[prop.name]['_id']).read_bytes()
    return commands.cast_backend_to_python(context, prop, backend, data)


@commands.getfile.register()
def getfile(
    context: Context,
    prop: Property,
    dtype: File,
    backend: FileSystem,
    *,
    data: FileObjectData,
) -> Optional[Path]:
    filename = data['_id']
    if filename is None:
        return None
    else:
        return dtype.backend.path / filename



from spinta import commands
from spinta.components import Context
from spinta.datasets.backends.sql.components import Sql


@commands.bootstrap.register(Context, Sql)
def bootstrap(context: Context, backend: Sql):
    pass



import json
from io import TextIOWrapper
from typing import cast

import itertools
from starlette.datastructures import UploadFile
from starlette.requests import Request

from spinta import commands
from spinta import exceptions
from spinta.formats.components import Format
from spinta.components import Action
from spinta.components import Context
from spinta.components import Node
from spinta.components import Store
from spinta.components import UrlParams
from spinta.exceptions import BaseError
from spinta.exceptions import NoBackendConfigured
from spinta.exceptions import error_response
from spinta.renderer import render


async def _check_post(context: Context, request: Request, params: UrlParams):
    form = await request.form()
    field = cast(UploadFile, form['manifest'])
    filename = field.filename

    # ---8<----
    # FIXME: https://github.com/pallets/werkzeug/issues/1344
    field.file.readable = field.file._file.readable
    field.file.writable = field.file._file.writable
    field.file.seekable = field.file._file.seekable
    # --->8----

    # XXX: This is a quick and dirty hack, uploaded manifest file should be
    #      loaded as any other manifest. Here only tabluar manifest loading is
    #      hardcoded.
    from spinta.core.config import RawConfig
    from spinta.manifests.helpers import clone_manifest
    from spinta.manifests.helpers import detect_manifest_from_path
    from spinta.manifests.helpers import load_manifest_nodes
    from spinta.manifests.tabular.components import TabularManifest
    from spinta.manifests.tabular.helpers import read_tabular_manifest
    rc: RawConfig = context.get('rc')
    Manifest_ = detect_manifest_from_path(rc, filename)
    if issubclass(Manifest_, TabularManifest):
        schemas = read_tabular_manifest(
            Manifest_.format,
            path=filename,
            file=TextIOWrapper(field.file, encoding='utf-8'),
        )
    else:
        return {
            'status': 'error',
            'errors': [{
                'type': 'manifest',
                'code': 'ValueError',
                'message': (
                    "Can't detect manifest type from given path {filename!r}."
                ),
            }],
        }

    try:
        manifest = clone_manifest(context)
        load_manifest_nodes(context, manifest, schemas)
        commands.link(context, manifest)
        commands.check(context, manifest)
    except BaseError as e:
        return {
            'status': 'error',
            'errors': [error_response(e)],
        }
    else:
        return {
            'status': 'OK',
        }


async def _check(context: Context, request: Request, params: UrlParams):
    commands.authorize(context, Action.CHECK, params.model)

    if request.method == 'POST':
        data = await _check_post(context, request, params)
    else:
        data = None

    return render(
        context,
        request,
        params.model,
        params,
        data,
        action=Action.CHECK,
    )


async def create_http_response(
    context: Context,
    params: UrlParams,
    request: Request,
):
    store: Store = context.get('store')
    manifest = store.manifest

    if manifest.backend is None:
        raise NoBackendConfigured(manifest)

    if params.action == Action.CHECK:
        return await _check(context, request, params)

    if request.method in ('GET', 'HEAD'):
        context.attach('transaction', manifest.backend.transaction)

        if params.changes:
            _enforce_limit(context, params)
            return await commands.changes(
                context,
                params.model,
                request,
                action=params.action,
                params=params,
            )

        elif params.pk:
            model = params.model
            action = params.action

            return await commands.getone(
                context,
                request,
                model,
                action=action,
                params=params,
            )

        else:
            _enforce_limit(context, params)
            action = params.action
            model = params.model
            backend = model.backend

            if backend is not None:
                # Namespace nodes do not have backend.
                context.attach(f'transaction.{backend.name}', backend.begin)

            if model.keymap:
                context.attach(
                    f'keymap.{model.keymap.name}',
                    lambda: model.keymap,
                )

            return await commands.getall(
                context,
                model,
                request,
                action=action,
                params=params,
            )

    elif request.method == 'DELETE' and params.action == Action.WIPE:
        if params.pk:
            raise NotImplementedError
        else:
            context.attach(
                'transaction',
                manifest.backend.transaction,
                write=True,
            )
            return await commands.wipe(
                context,
                request,
                params.model,
                params.model.backend,
                action=params.action,
                params=params,
            )

    else:
        context.attach('transaction', manifest.backend.transaction, write=True)
        action = params.action
        if params.prop and params.propref:
            return await commands.push(
                context,
                request,
                params.prop.dtype,
                params.model.backend,
                action=action,
                params=params,
            )
        elif params.prop:
            return await commands.push(
                context,
                request,
                params.prop.dtype,
                params.prop.dtype.backend,
                action=action,
                params=params,
            )
        else:
            return await commands.push(
                context,
                request,
                params.model,
                params.model.backend,
                action=action,
                params=params,
            )


def _enforce_limit(context: Context, params: UrlParams):
    config = context.get('config')
    fmt: Format = config.exporters[params.format]
    # XXX: I think this is not the best way to enforce limit, maybe simply
    #      an error should be raised?
    # XXX: Max resource count should be configurable.
    if not fmt.streamable and (params.limit is None or params.limit > 100):
        params.limit = params.limit_enforced_to + 1
        params.limit_enforced = True


def peek_and_stream(stream):
    peek = list(itertools.islice(stream, 2))

    def _iter():
        for data in itertools.chain(peek, stream):
            yield data

    return _iter()


async def aiter(stream):
    for data in stream:
        yield data


async def get_request_data(node: Node, request: Request):
    ct = request.headers.get('content-type')
    if ct != 'application/json':
        raise exceptions.UnknownContentType(
            node,
            content_type=ct,
            supported_content_types=['application/json'],
        )

    try:
        data = await request.json()
    except json.decoder.JSONDecodeError as e:
        raise exceptions.JSONError(node, error=str(e))

    return data



from typing import overload
from typing import Iterator

from spinta.typing import ObjectData
from spinta import commands
from spinta.core.ufuncs import Expr
from spinta.components import Context
from spinta.components import Model
from spinta.exceptions import NotFoundError
from spinta.exceptions import ItemDoesNotExist
from spinta.backends.postgresql.components import PostgreSQL
from spinta.utils.nestedstruct import flat_dicts_to_nested
from spinta.backends.postgresql.commands.query import PgQueryBuilder


@overload
@commands.getone.register(Context, Model, PostgreSQL)
def getone(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    id_: str,
) -> ObjectData:
    connection = context.get('transaction').connection
    table = backend.get_table(model)
    try:
        result = backend.get(connection, table, table.c._id == id_)
    except NotFoundError:
        raise ItemDoesNotExist(model, id=id_)
    data = flat_dicts_to_nested(dict(result))
    data['_type'] = model.model_type()
    return commands.cast_backend_to_python(context, model, backend, data)


@commands.getall.register(Context, Model, PostgreSQL)
def getall(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    query: Expr = None,
) -> Iterator[ObjectData]:
    assert isinstance(query, (Expr, type(None))), query
    connection = context.get('transaction').connection

    builder = PgQueryBuilder(context)
    builder.update(model=model)
    table = backend.get_table(model)
    env = builder.init(backend, table)
    expr = env.resolve(query)
    where = env.execute(expr)
    qry = env.build(where)

    conn = connection.execution_options(stream_results=True)
    result = conn.execute(qry)

    for row in result:
        row = flat_dicts_to_nested(dict(row))
        row = {
            '_type': model.model_type(),
            **row,
        }
        row = commands.cast_backend_to_python(context, model, backend, row)
        yield row
